import pytest
import os

lib_installed = os.environ.get('KIVY_EVENTLOOP_TEST_INSTALLED', None)
event_loop = os.environ.get('KIVY_EVENTLOOP', 'asyncio')
if lib_installed is not None and lib_installed != event_loop:
    pytestmark = pytest.mark.skip(
        'Tests are run only when event loop matches async library installed')
elif event_loop == 'asyncio':
    pytestmark = pytest.mark.asyncio
else:
    pytestmark = pytest.mark.trio


def button_app():
    from kivy.app import App
    from kivy.uix.togglebutton import ToggleButton

    class TestApp(App):
        def build(self):
            return ToggleButton(text='Hello, World!')

    return TestApp()


async def assert_app_working(app):
    root = app.app.root
    assert root.text == 'Hello, World!'
    assert root.state == 'normal'

    async for _ in app.do_touch_down_up(widget=root, widget_jitter=True):
        pass

    assert root.state == 'down'


async def test_app_call_cls(async_kivy_app):
    await async_kivy_app(button_app)
    await assert_app_working(async_kivy_app)


@pytest.mark.parametrize(
    'async_kivy_app', [{'app_cls': button_app}], indirect=True)
async def test_app_param_cls(async_kivy_app):
    await assert_app_working(async_kivy_app)


@pytest.mark.parametrize(
    'async_kivy_app', [{'kwargs': {'height': 400}}], indirect=True)
async def test_app_param_height(async_kivy_app):
    await async_kivy_app(button_app)
    await assert_app_working(async_kivy_app)
    assert async_kivy_app.app.root.height == 400



"""Tools
========


"""

__all__ = ('exhaust', )


async def exhaust(async_it):
    """Iterates through all the elements of the iterator until done.
    It does nothing with the elements.
    """
    async for _ in async_it:
        pass



"""Input
========

Input classes that can be used to simulate deice (e.g. mouse) input.
"""
from kivy.tests import UnitTestTouch

__all__ = ('AsyncUnitTestTouch', )


class AsyncUnitTestTouch(UnitTestTouch):

    def __init__(self, *largs, **kwargs):
        self.grab_exclusive_class = None
        self.is_touch = True
        super(AsyncUnitTestTouch, self).__init__(*largs, **kwargs)

    def touch_down(self, *args):
        self.eventloop._dispatch_input("begin", self)

    def touch_move(self, x, y):
        win = self.eventloop.window
        self.move({
            "x": x / float(win.width),
            "y": y / float(win.height)
        })
        self.eventloop._dispatch_input("update", self)

    def touch_up(self, *args):
        self.eventloop._dispatch_input("end", self)



import pytest
import os

lib_installed = os.environ.get('KIVY_EVENTLOOP_TEST_INSTALLED', None)
event_loop = os.environ.get('KIVY_EVENTLOOP', 'asyncio')
if lib_installed is not None and lib_installed != event_loop:
    pytestmark = pytest.mark.skip(
        'Tests are run only when event loop matches async library installed')
elif event_loop == 'asyncio':
    pytestmark = pytest.mark.asyncio
else:
    pytestmark = pytest.mark.trio


async def test_touch_down_up(async_kivy_app):
    def button_app():
        from kivy.app import App
        from kivy.uix.togglebutton import ToggleButton

        class TestApp(App):
            def build(self):
                return ToggleButton(text='Hello, World!')

        return TestApp()

    await async_kivy_app(button_app)

    root = async_kivy_app.app.root
    assert root.text == 'Hello, World!'
    assert root.state == 'normal'

    async for _ in async_kivy_app.do_touch_down_up(
            widget=root, widget_jitter=True):
        pass

    assert root.state == 'down'


@pytest.mark.parametrize(
    'async_kivy_app', [{'kwargs': {'height': 200, 'width': 200}}],
    indirect=True)
async def test_drag_scroll_view_pixels(async_kivy_app):
    def create_app():
        from kivy.app import App
        from kivy.lang import Builder
        from textwrap import dedent

        kv = '''
        ScrollView:
            scroll_type: ['bars']
            bar_width: '15dp'
            BoxLayout:
                width: 400
                size_hint_x: None
                Widget:
                    canvas:
                        Color:
                            rgba: 1, 0, 0, 1
                        Rectangle:
                            pos: self.pos
                            size: self.size
                Widget:
                    canvas:
                        Color:
                            rgba: 0, 1, 0, 1
                        Rectangle:
                            pos: self.pos
                            size: self.size
        '''

        class TestApp(App):
            def build(self):
                return Builder.load_string(dedent(kv))

        return TestApp()

    await async_kivy_app(create_app)

    root = async_kivy_app.app.root
    (r, g, b, a), = async_kivy_app.get_widget_pos_pixel(root, [(100, 100)])
    assert r == 255
    assert not g
    assert not b
    assert a == 255

    # drag scrollview to show second widget
    async for _ in async_kivy_app.do_touch_drag(pos=(5, 7), dx=185):
        pass

    (r, g, b, a), = async_kivy_app.get_widget_pos_pixel(root, [(100, 100)])
    assert not r
    assert g == 255
    assert not b
    assert a == 255


async def test_text_app(async_kivy_app):
    def create_app():
        from kivy.app import App
        from kivy.uix.textinput import TextInput

        class TestApp(App):
            def build(self):
                return TextInput()

        return TestApp()

    await async_kivy_app(create_app)
    root = async_kivy_app.app.root

    assert root.text == ''

    # activate widget
    async for state, touch_pos in async_kivy_app.do_touch_down_up(widget=root):
        pass

    async for state, value in async_kivy_app.do_keyboard_key(
            key='A', num_press=4):
        pass
    async for state, value in async_kivy_app.do_keyboard_key(
            key='q', num_press=3):
        pass

    assert root.text == 'AAAAqqq'



"""App
======

The wrapper app to help test Kivy apps.

"""

import random
import time
import math
import os

from pytest_kivy.resolver import WidgetResolver

__all__ = ('AsyncUnitApp', )


class AsyncUnitApp:
    """Wrapper app that provides methods to test parts of a Kivy App
    asynchronously.
    """

    app_has_started = False

    app_has_stopped = False

    app = None

    _context = None

    _nursery = None

    _event_loop = None

    width = 320

    height = 240

    async_lib = os.environ.get('KIVY_EVENTLOOP', 'asyncio')

    def __init__(
            self, nursery=None, event_loop=None, width=320, height=240,
            async_lib=os.environ.get('KIVY_EVENTLOOP', 'asyncio')):
        super().__init__()
        self._nursery = nursery
        self._event_loop = event_loop
        self.width = width
        self.height = height
        self.async_lib = async_lib

    def __getattr__(self, item):
        return getattr(self.app, item)

    async def __aenter__(self):
        from kivy.config import Config
        Config.set('graphics', 'width', str(self.width))
        Config.set('graphics', 'height', str(self.height))
        for items in Config.items('input'):
            Config.remove_option('input', items[0])

        from kivy.core.window import Window
        from kivy.context import Context
        from kivy.clock import ClockBase
        from kivy.factory import FactoryBase, Factory
        from kivy.lang.builder import BuilderBase, Builder

        kivy_eventloop = self.async_lib
        if kivy_eventloop == 'asyncio':
            try:
                import pytest_asyncio
            except ImportError as e:
                raise TypeError(
                    'KIVY_EVENTLOOP is asyncio but "pytest_asyncio" is not '
                    'installed') from e
            async_lib = 'asyncio'
        elif kivy_eventloop == 'trio':
            try:
                import pytest_trio
            except ImportError as e:
                raise TypeError(
                    'KIVY_EVENTLOOP is trio but "pytest_trio" is not '
                    'installed') from e
            async_lib = 'trio'
        else:
            raise TypeError(f'unknown event loop {kivy_eventloop}')

        self._context = context = Context(init=False)
        context['Clock'] = ClockBase(async_lib=async_lib)
        # have to make sure all global kv files are loaded before this because
        # globally read kv files (e.g. on module import) will not be loaded
        # again in the new builder, except if manually loaded, which we don't
        # context['Factory'] = FactoryBase.create_from(Factory)
        # context['Builder'] = BuilderBase.create_from(Builder)
        context.push()

        Window.create_window()
        Window.register()
        Window.initialized = True
        Window.canvas.clear()
        Window.size = self.width, self.height

        from kivy.clock import Clock
        Clock._max_fps = 0
        Clock.init_async_lib(async_lib)
        return self

    async def __call__(self, app_cls):
        self.app = app = app_cls()

        def started_app(*largs):
            self.app_has_started = True
        app.fbind('on_start', started_app)

        def stopped_app(*largs):
            self.app_has_stopped = True
        app.fbind('on_stop', stopped_app)

        if self.async_lib == 'asyncio':
            self._event_loop.create_task(app.async_run())
        else:
            self._nursery.start_soon(app.async_run)

        ts = time.perf_counter()
        while not self.app_has_started:
            await self.async_sleep(.1)
            if time.perf_counter() - ts >= 120:
                raise TimeoutError()

        await self.wait_clock_frames(5)
        return app

    async def wait_stop_app(self):
        from kivy.base import stopTouchApp
        stopTouchApp()

        ts = time.perf_counter()
        while not self.app_has_stopped:
            await self.async_sleep(.1)
            if time.perf_counter() - ts >= 60:
                raise TimeoutError()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        from kivy.core.window import Window
        from kivy.animation import Animation
        from kivy.base import stopTouchApp
        from kivy.logger import LoggerHistory

        stopTouchApp()
        for anim in list(Animation._instances):
            anim._unregister()
        for child in Window.children[:]:
            Window.remove_widget(child)

        self._context.pop()
        self._context = None
        LoggerHistory.clear_history()

    async def async_sleep(self, delay):
        from kivy.clock import Clock
        await Clock._async_lib.sleep(delay)

    def resolve_widget(self, base_widget=None):
        if base_widget is None:
            from kivy.core.window import Window
            base_widget = Window
        return WidgetResolver(base_widget=base_widget)

    async def wait_clock_frames(self, n, sleep_time=1 / 60.):
        from kivy.clock import Clock
        frames_start = Clock.frames
        while Clock.frames < frames_start + n:
            await self.async_sleep(sleep_time)

    def get_widget_pos_pixel(self, widget, positions):
        from kivy.graphics import Fbo, ClearColor, ClearBuffers

        canvas_parent_index = -2
        if widget.parent is not None:
            canvas_parent_index = widget.parent.canvas.indexof(widget.canvas)
            if canvas_parent_index > -1:
                widget.parent.canvas.remove(widget.canvas)

        w, h = int(widget.width), int(widget.height)
        fbo = Fbo(size=(w, h), with_stencilbuffer=True)

        with fbo:
            ClearColor(0, 0, 0, 0)
            ClearBuffers()

        fbo.add(widget.canvas)
        fbo.draw()
        pixels = fbo.pixels
        fbo.remove(widget.canvas)

        if widget.parent is not None and canvas_parent_index > -1:
            widget.parent.canvas.insert(canvas_parent_index, widget.canvas)

        values = []
        for x, y in positions:
            x = int(x)
            y = int(y)
            i = y * w * 4 + x * 4
            values.append(tuple(pixels[i:i + 4]))

        return values

    async def do_touch_down_up(
            self, pos=None, widget=None, duration=.2, pos_jitter=None,
            widget_jitter=False, jitter_dt=1 / 15., end_on_pos=False):
        from pytest_kivy.input import AsyncUnitTestTouch
        if widget is None:
            x, y = pos
        else:
            if pos is None:
                x, y = widget.to_window(*widget.center)
            else:
                x, y = widget.to_window(*pos, initial=False)
        touch = AsyncUnitTestTouch(x, y)

        ts = time.perf_counter()
        touch.touch_down()
        await self.wait_clock_frames(1)
        yield 'down', touch.pos

        if not pos_jitter and not widget_jitter:
            await self.async_sleep(duration)
            touch.touch_up()
            await self.wait_clock_frames(1)
            yield 'up', touch.pos

            return

        moved = False
        if pos_jitter:
            dx, dy = pos_jitter
        else:
            dx = widget.width / 2.
            dy = widget.height / 2.

        while time.perf_counter() - ts < duration:
            moved = True
            await self.async_sleep(jitter_dt)

            touch.touch_move(
                x + (random.random() * 2 - 1) * dx,
                y + (random.random() * 2 - 1) * dy
            )
            await self.wait_clock_frames(1)
            yield 'move', touch.pos

        if end_on_pos and moved:
            touch.touch_move(x, y)
            await self.wait_clock_frames(1)
            yield 'move', touch.pos

        touch.touch_up()
        await self.wait_clock_frames(1)
        yield 'up', touch.pos

    async def do_touch_drag(
            self, pos=None, widget=None,
            widget_loc=('center_x', 'center_y'), dx=0, dy=0,
            target_pos=None, target_widget=None, target_widget_offset=(0, 0),
            target_widget_loc=('center_x', 'center_y'), long_press=0,
            duration=.2, drag_n=5):
        """Initiates a touch down, followed by some dragging to a target
        location, ending with a touch up.

        `origin`: These parameters specify where the drag starts.

        - If ``widget`` is None, it starts at ``pos`` (in window coordinates).
          If ``dx``/``dy`` is used, it is in the window coordinate system also.
        - If ``pos`` is None, it starts on the ``widget`` as specified by
          ``widget_loc``. If ``dx``/``dy`` is used, it is in the ``widget``
          coordinate system.
        - If neither is None, it starts at ``pos``, but in the ``widget``'s
          coordinate system (:meth:`~kivy.uix.widget.Widget.to_window` is used
          on it). If ``dx``/``dy`` is used, it is in the ``widget``
          coordinate system.

        `target`: These parameters specify where the drag ends.

        - If ``target_pos`` and ``target_widget`` is None, then ``dx`` and
          ``dy`` is used relative to the position where the drag started.
        - If ``target_widget`` is None, it ends at ``target_pos``
          (in window coordinates).
        - If ``target_pos`` is None, it ends on the ``target_widget`` as
          specified by ``target_widget_loc``. ``target_widget_offset``, is an
          additional ``(x, y)`` offset relative to ``target_widget_loc``.
        - If neither is None, it starts at ``target_pos``, but in the
          ``target_widget``'s coordinate system
          (:meth:`~kivy.uix.widget.Widget.to_window` is used on it).

        When ``widget`` and/or ``target_widget`` are specified, ``widget_loc``
        and ``target_widget_loc``, respectively, indicate where on the widget
        the drag starts/ends. It is a a tuple with property names of the widget
        to loop up to get the value. The default is
        ``('center_x', 'center_y')`` so the drag would start/end in the
        widget's center.
        """
        from pytest_kivy.input import AsyncUnitTestTouch
        if widget is None:
            x, y = pos
            tx, ty = x + dx, y + dy
        else:
            if pos is None:
                w_x = getattr(widget, widget_loc[0])
                w_y = getattr(widget, widget_loc[1])
                x, y = widget.to_window(w_x, w_y)
                tx, ty = widget.to_window(w_x + dx, w_y + dy)
            else:
                x, y = widget.to_window(*pos, initial=False)
                tx, ty = widget.to_window(
                    pos[0] + dx, pos[1] + dy, initial=False)

        if target_pos is not None:
            if target_widget is None:
                tx, ty = target_pos
            else:
                tx, ty = target_pos = target_widget.to_window(
                    *target_pos, initial=False)
        elif target_widget is not None:
            x_off, y_off = target_widget_offset
            w_x = getattr(target_widget, target_widget_loc[0]) + x_off
            w_y = getattr(target_widget, target_widget_loc[1]) + y_off
            tx, ty = target_pos = target_widget.to_window(w_x, w_y)
        else:
            target_pos = tx, ty

        touch = AsyncUnitTestTouch(x, y)

        touch.touch_down()
        await self.wait_clock_frames(1)
        if long_press:
            await self.async_sleep(long_press)
        yield 'down', touch.pos

        dx = (tx - x) / drag_n
        dy = (ty - y) / drag_n

        ts0 = time.perf_counter()
        for i in range(drag_n):
            await self.async_sleep(
                max(0., duration - (time.perf_counter() - ts0)) / (drag_n - i))

            touch.touch_move(x + (i + 1) * dx, y + (i + 1) * dy)
            await self.wait_clock_frames(1)
            yield 'move', touch.pos

        if touch.pos != target_pos:
            touch.touch_move(*target_pos)
            await self.wait_clock_frames(1)
            yield 'move', touch.pos

        touch.touch_up()
        await self.wait_clock_frames(1)
        yield 'up', touch.pos

    async def do_touch_drag_follow(
            self, pos=None, widget=None,
            widget_loc=('center_x', 'center_y'),
            target_pos=None, target_widget=None, target_widget_offset=(0, 0),
            target_widget_loc=('center_x', 'center_y'), long_press=0,
            duration=.2, drag_n=5, max_n=25):
        """Very similar to :meth:`do_touch_drag`, except it follows the target
        widget, even if the target widget moves as a result of the drag, the
        drag will follow it until it's on the target widget.

        `origin`: These parameters specify where the drag starts.

        - If ``widget`` is None, it starts at ``pos`` (in window coordinates).
        - If ``pos`` is None, it starts on the ``widget`` as specified by
          ``widget_loc``.
        - If neither is None, it starts at ``pos``, but in the ``widget``'s
          coordinate system (:meth:`~kivy.uix.widget.Widget.to_window` is used
          on it).

        `target`: These parameters specify where the drag ends.

        - If ``target_pos`` is None, it ends on the ``target_widget`` as
          specified by ``target_widget_loc``. ``target_widget_offset``, is an
          additional ``(x, y)`` offset relative to ``target_widget_loc``.
        - If ``target_pos`` is not None, it starts at ``target_pos``, but in
          the ``target_widget``'s coordinate system
          (:meth:`~kivy.uix.widget.Widget.to_window` is used on it).

        When ``widget`` and/or ``target_widget`` are specified, ``widget_loc``
        and ``target_widget_loc``, respectively, indicate where on the widget
        the drag starts/ends. It is a a tuple with property names of the widget
        to loop up to get the value. The default is
        ``('center_x', 'center_y')`` so the drag would start/end in the
        widget's center.
        """
        from pytest_kivy.input import AsyncUnitTestTouch
        if widget is None:
            x, y = pos
        else:
            if pos is None:
                w_x = getattr(widget, widget_loc[0])
                w_y = getattr(widget, widget_loc[1])
                x, y = widget.to_window(w_x, w_y)
            else:
                x, y = widget.to_window(*pos, initial=False)

        if target_widget is None:
            raise ValueError('target_widget must be specified')

        def get_target():
            if target_pos is not None:
                return target_widget.to_window(*target_pos, initial=False)
            else:
                x_off, y_off = target_widget_offset
                wt_x = getattr(target_widget, target_widget_loc[0]) + x_off
                wt_y = getattr(target_widget, target_widget_loc[1]) + y_off
                return target_widget.to_window(wt_x, wt_y)

        touch = AsyncUnitTestTouch(x, y)

        touch.touch_down()
        await self.wait_clock_frames(1)
        if long_press:
            await self.async_sleep(long_press)
        yield 'down', touch.pos

        ts0 = time.perf_counter()
        tx, ty = get_target()
        i = 0
        while not (math.isclose(touch.x, tx) and math.isclose(touch.y, ty)):
            if i >= max_n:
                raise Exception(
                    'Exceeded the maximum number of iterations, '
                    'but {} != {}'.format(touch.pos, (tx, ty)))

            rem_i = max(1, drag_n - i)
            rem_t = max(0., duration - (time.perf_counter() - ts0)) / rem_i
            i += 1
            await self.async_sleep(rem_t)

            x, y = touch.pos
            touch.touch_move(x + (tx - x) / rem_i, y + (ty - y) / rem_i)
            await self.wait_clock_frames(1)
            yield 'move', touch.pos
            tx, ty = get_target()

        touch.touch_up()
        await self.wait_clock_frames(1)
        yield 'up', touch.pos

    async def do_touch_drag_path(
            self, path, axis_widget=None, long_press=0, duration=.2):
        """Drags the touch along the specified path.

        :parameters:

            `path`: list
                A list of position tuples the touch will follow. The first
                item is used for the touch down and the rest for the move.
            `axis_widget`: a Widget
                If None, the path coordinates is in window coordinates,
                otherwise, we will first transform the path coordinates
                to window coordinates using
                :meth:`~kivy.uix.widget.Widget.to_window` of the specified
                widget.
        """
        from pytest_kivy.input import AsyncUnitTestTouch
        if axis_widget is not None:
            path = [axis_widget.to_window(*p, initial=False) for p in path]
        x, y = path[0]
        path = path[1:]

        touch = AsyncUnitTestTouch(x, y)

        touch.touch_down()
        await self.wait_clock_frames(1)
        if long_press:
            await self.async_sleep(long_press)
        yield 'down', touch.pos

        ts0 = time.perf_counter()
        n = len(path)
        for i, (x2, y2) in enumerate(path):
            await self.async_sleep(
                max(0., duration - (time.perf_counter() - ts0)) / (n - i))

            touch.touch_move(x2, y2)
            await self.wait_clock_frames(1)
            yield 'move', touch.pos

        touch.touch_up()
        await self.wait_clock_frames(1)
        yield 'up', touch.pos

    async def do_keyboard_key(
            self, key, modifiers=(), duration=.05, num_press=1):
        from kivy.core.window import Window
        if key == ' ':
            key = 'spacebar'
        key_lower = key.lower()
        key_code = Window._system_keyboard.string_to_keycode(key_lower)

        known_modifiers = {'shift', 'alt', 'ctrl', 'meta'}
        if set(modifiers) - known_modifiers:
            raise ValueError('Unknown modifiers "{}"'.
                             format(set(modifiers) - known_modifiers))

        special_keys = {
            27: 'escape',
            9: 'tab',
            8: 'backspace',
            13: 'enter',
            127: 'del',
            271: 'enter',
            273: 'up',
            274: 'down',
            275: 'right',
            276: 'left',
            278: 'home',
            279: 'end',
            280: 'pgup',
            281: 'pgdown',
            300: 'numlock',
            301: 'capslock',
            145: 'screenlock',
        }

        text = None
        try:
            text = chr(key_code)
            if key_lower != key:
                text = key
        except ValueError:
            pass

        dt = duration / num_press
        for i in range(num_press):
            await self.async_sleep(dt)

            Window.dispatch('on_key_down', key_code, 0, text, modifiers)
            if (key not in known_modifiers and
                    key_code not in special_keys and
                    not (known_modifiers & set(modifiers))):
                Window.dispatch('on_textinput', text)

            await self.wait_clock_frames(1)
            yield 'down', (key, key_code, 0, text, modifiers)

        Window.dispatch('on_key_up', key_code, 0)
        await self.wait_clock_frames(1)
        yield 'up', (key, key_code, 0, text, modifiers)



"""Resolver
===========

Resolves individual widgets, given a root widget to help locate and test
a specific widget in the widget tree.

"""

from collections import deque

__all__ = ('WidgetResolver', )

_unique_value = object


class WidgetResolver:
    """It assumes that the widget tree strictly forms a DAG.
    """

    base_widget = None

    matched_widget = None

    _kwargs_filter = {}

    _funcs_filter = []

    def __init__(self, base_widget, **kwargs):
        self.base_widget = base_widget
        self._kwargs_filter = {}
        self._funcs_filter = []
        super(WidgetResolver, self).__init__(**kwargs)

    def __call__(self):
        if self.matched_widget is not None:
            return self.matched_widget

        if not self._kwargs_filter and not self._funcs_filter:
            return self.base_widget
        return None

    def match(self, **kwargs_filter):
        self._kwargs_filter.update(kwargs_filter)

    def match_funcs(self, funcs_filter=()):
        self._funcs_filter.extend(funcs_filter)

    def check_widget(self, widget):
        if not all(func(widget) for func in self._funcs_filter):
            return False

        for attr, val in self._kwargs_filter.items():
            if getattr(widget, attr, _unique_value) != val:
                return False

        return True

    def not_found(self, op):
        raise ValueError(
            'Cannot find widget matching <{}, {}> starting from base '
            'widget "{}" doing "{}" traversal'.format(
                self._kwargs_filter, self._funcs_filter, self.base_widget, op))

    def down(self, **kwargs_filter):
        self.match(**kwargs_filter)
        check = self.check_widget

        fifo = deque([self.base_widget])
        while fifo:
            widget = fifo.popleft()
            if check(widget):
                return WidgetResolver(base_widget=widget)

            fifo.extend(widget.children)

        self.not_found('down')

    def up(self, **kwargs_filter):
        self.match(**kwargs_filter)
        check = self.check_widget

        parent = self.base_widget
        while parent is not None:
            if check(parent):
                return WidgetResolver(base_widget=parent)

            new_parent = parent.parent
            # Window is its own parent oO
            if new_parent is parent:
                break
            parent = new_parent

        self.not_found('up')

    def family_up(self, **kwargs_filter):
        self.match(**kwargs_filter)
        check = self.check_widget

        base_widget = self.base_widget
        already_checked_base = None
        while base_widget is not None:
            fifo = deque([base_widget])
            while fifo:
                widget = fifo.popleft()
                # don't check the child we checked before moving up
                if widget is already_checked_base:
                    continue

                if check(widget):
                    return WidgetResolver(base_widget=widget)

                fifo.extend(widget.children)

            already_checked_base = base_widget
            new_base_widget = base_widget.parent
            # Window is its own parent oO
            if new_base_widget is base_widget:
                break
            base_widget = new_base_widget

        self.not_found('family_up')



"""pytest-trio
==============

Provides pytest fixtures to more easily test Kivy GUI apps.
"""
from pytest_kivy._version import __version__

import pytest

pytest.register_assert_rewrite("pytest_kivy.resolver")
pytest.register_assert_rewrite("pytest_kivy.app")



"""Plugin
=========

The fixtures provided by pytest-kivy.
"""
import pytest
import weakref
from typing import Tuple, Type, Optional, Callable
import gc
import logging
from os import environ

from pytest_kivy.app import AsyncUnitApp

__all__ = ('trio_kivy_app', 'asyncio_kivy_app', 'async_kivy_app')

#: NOTE: Kivy cannot be imported before or while the plugin is imported or
# configured as that leads to pytest issues.

environ['KIVY_USE_DEFAULTCONFIG'] = '1'

_async_lib = environ.get('KIVY_EVENTLOOP', 'asyncio')
if _async_lib == 'asyncio':
    @pytest.fixture
    async def _nursery():
        return None

    @pytest.fixture
    async def _event_loop(event_loop):
        return event_loop
elif _async_lib == 'trio':
    @pytest.fixture
    async def _nursery(nursery):
        return nursery

    @pytest.fixture
    async def _event_loop():
        return None
else:
    raise TypeError(f'unknown event loop {_async_lib}')


def pytest_addoption(parser):
    group = parser.getgroup("kivy")
    group.addoption(
        "--kivy-app-release",
        action="store_true",
        default=False,
        help='Whether to check after each test if the app is released and no '
             'references are kept to the app preventing it from being garbage '
             'collected.',
    )
    group.addoption(
        "--kivy-app-release-end",
        action="store_true",
        default=False,
        help='Whether to check at the end of all tests if all of the test apps'
             'were released and no references were kept to the app preventing'
             'them from being garbage collected.',
    )


@pytest.fixture(scope='session')
def _app_release_list():
    apps = []

    yield apps

    gc.collect()
    alive_apps = []
    for i, (app, request) in enumerate(apps[1:-1]):
        app = app()
        request = request()
        if request is None:
            request = '<dead request>'

        if app is not None:
            alive_apps.append((app, request))
            logging.error(
                'Memory leak: failed to release app for test ' + repr(request))

    assert not alive_apps, 'Memory leak: failed to release all apps'


@pytest.fixture
def _app_release():
    app = []

    yield app

    gc.collect()

    if not app:
        return

    app, request = app[0]
    app = app()
    request = request()
    if request is None:
        request = '<dead request>'

    assert app is None, \
        f'Memory leak: failed to release app for test {request!r}'


def _get_request_config(
        request, _app_release_list, _app_release
) -> Tuple[Type[AsyncUnitApp], dict, Optional[Callable], list]:
    opts = getattr(request, 'param', {})
    cls = opts.get('cls', AsyncUnitApp)
    kwargs = opts.get('kwargs', {})
    app_cls = opts.get('app_cls', None)

    app_list = None
    if request.config.getoption("kivy_app_release"):
        app_list = _app_release
    elif request.config.getoption("kivy_app_release_end"):
        app_list = _app_release_list
    return cls, kwargs, app_cls, app_list


@pytest.fixture
async def trio_kivy_app(
        request, nursery, _app_release_list, _app_release
) -> AsyncUnitApp:
    """Fixture yielding a :class:`~pytest_kivy.app.AsyncUnitApp` using
    explicitly trio as backend for the async library.

    pytest-trio and trio must be installed, and ``trio_mode = true`` must be
    set in pytest.ini.
    """
    cls, kwargs, app_cls, app_list = _get_request_config(
        request, _app_release_list, _app_release)

    async with cls(nursery=nursery, async_lib='trio', **kwargs) as app:
        if app_list is not None:
            app_list.append((weakref.ref(app), weakref.ref(request)))

        if app_cls is not None:
            await app(app_cls)
        yield app
        await app.wait_stop_app()


@pytest.fixture
async def asyncio_kivy_app(
        request, event_loop, _app_release_list, _app_release) -> AsyncUnitApp:
    """Fixture yielding a :class:`~pytest_kivy.app.AsyncUnitApp` using
    explicitly asyncio as backend for the async library.

    pytest-asyncio must be installed.
    """
    cls, kwargs, app_cls, app_list = _get_request_config(
        request, _app_release_list, _app_release)

    async with cls(
            event_loop=event_loop, async_lib='asyncio', **kwargs) as app:
        if app_list is not None:
            app_list.append((weakref.ref(app), weakref.ref(request)))

        if app_cls is not None:
            await app(app_cls)
        yield app
        await app.wait_stop_app()


@pytest.fixture
async def async_kivy_app(
        request, _app_release_list, _app_release, _nursery, _event_loop
) -> AsyncUnitApp:
    """Fixture yielding a :class:`~pytest_kivy.app.AsyncUnitApp` using
    trio or asyncio as backend for the async library, depending on
    KIVY_EVENTLOOP.

    If using trio, pytest-trio and trio must be installed, and
    ``trio_mode = true`` must be set in pytest.ini. If using asyncio,
    pytest-asyncio must be installed.
    """
    cls, kwargs, app_cls, app_list = _get_request_config(
        request, _app_release_list, _app_release)

    async with cls(
            nursery=_nursery, event_loop=_event_loop, async_lib=_async_lib,
            **kwargs) as app:
        if app_list is not None:
            app_list.append((weakref.ref(app), weakref.ref(request)))

        if app_cls is not None:
            await app(app_cls)
        yield app
        await app.wait_stop_app()



from setuptools import setup, find_packages
from io import open
from os.path import dirname, join, abspath

here = abspath(dirname(__file__))

# __version__ is imported by exec, but help linter not complain
__version__ = None
with open(join(here, 'pytest_kivy', '_version.py'), encoding="utf-8") as f:
    exec(f.read())


with open(join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

URL = 'https://github.com/matham/pytest-kivy'

setup(
    name='pytest-kivy',
    version=__version__,
    author='Matthew Einhorn',
    author_email='moiein2000@gmail.com',
    license='MIT',
    description='Kivy GUI tests fixtures using pytest',
    long_description=long_description,
    url=URL,
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Software Development :: Testing",
        "Framework :: Pytest",
    ],
    packages=find_packages(),
    entry_points={"pytest11": ["kivy = pytest_kivy.plugin"]},
    project_urls={
        'Bug Reports': URL + '/issues',
        'Source': URL,
    },
)



import os
import pytest

lib_installed = os.environ.get('KIVY_EVENTLOOP_TEST_INSTALLED', None)
event_loop = os.environ.get('KIVY_EVENTLOOP', None)
if lib_installed == 'asyncio' or lib_installed is None:
    pytestmark = pytest.mark.asyncio
else:
    pytestmark = pytest.mark.trio


def button_app():
    from kivy.app import App
    from kivy.uix.togglebutton import ToggleButton

    class TestApp(App):
        def build(self):
            return ToggleButton(text='Hello, World!')

    return TestApp()


async def assert_app_working(app):
    root = app.app.root
    assert root.text == 'Hello, World!'
    assert root.state == 'normal'

    async for _ in app.do_touch_down_up(widget=root, widget_jitter=True):
        pass

    assert root.state == 'down'


@pytest.mark.skipif(lib_installed != 'trio', reason='Need trio installed')
async def test_button_app_trio(trio_kivy_app):
    await trio_kivy_app(button_app)
    await assert_app_working(trio_kivy_app)


@pytest.mark.skipif(
    lib_installed != 'asyncio', reason='Need asyncio installed')
async def test_button_app_asyncio(asyncio_kivy_app):
    await asyncio_kivy_app(button_app)
    await assert_app_working(asyncio_kivy_app)


@pytest.mark.skipif(
    lib_installed != event_loop, reason='Installed must match event loop')
async def test_button_app_async(async_kivy_app):
    await async_kivy_app(button_app)
    await assert_app_working(async_kivy_app)
